"""
Utilities
==============
Rate limiter classes.

These are basically callables that when called register that a request was
issued.  Depending on how they are configured that may cause a pause or exception
if a rate limit has been exceeded.  Obviously it is up to the calling code to ensure
that these callables are invoked with every (successful?) call to the backend
API.  (There is probably a better way to hook these into the requests library
directly ... TBD.)

From the Strava docs:
  Strava API usage is limited on a per-application basis using a short term,
  15 minute, limit and a long term, daily, limit. The default rate limit allows
  600 requests every 15 minutes, with up to 30,000 requests per day.

  This limit allows applications to make 40 requests per minute for about
  half the day.
"""
from __future__ import annotations

import collections
import logging
import time
from datetime import datetime, timedelta
from logging import Logger
from typing import Callable, Deque, Literal, NamedTuple, NoReturn, TypedDict

import arrow

from stravalib import exc


class RequestRate(NamedTuple):
    """Tuple about the request usage and usage limit."""

    short_usage: int
    """15-minute usage"""

    long_usage: int
    """Daily usage"""

    short_limit: int
    """15-minutes limit"""

    long_limit: int
    """Daily limit"""


def get_rates_from_response_headers(
    headers: dict[str, str]
) -> RequestRate | None:
    """
    Returns a namedtuple with values for short - and long usage and limit rates found in provided HTTP response headers
    :param headers: HTTP response headers
    :type headers: dict
    :return: namedtuple with request rates or None if no rate-limit headers present in response.
    :rtype: Optional[RequestRate]
    """
    try:
        usage_rates = [int(v) for v in headers["X-RateLimit-Usage"].split(",")]
        limit_rates = [int(v) for v in headers["X-RateLimit-Limit"].split(",")]

        return RequestRate(
            short_usage=usage_rates[0],
            long_usage=usage_rates[1],
            short_limit=limit_rates[0],
            long_limit=limit_rates[1],
        )
    except KeyError:
        return None


def get_seconds_until_next_quarter(
    now: arrow.arrow.Arrow | None = None,
) -> int:
    """
    Returns the number of seconds until the next quarter of an hour. This is the short-term rate limit used by Strava.
    :param now: A (utc) timestamp
    :type now: arrow.arrow.Arrow
    :return: the number of seconds until the next quarter, as int
    """
    if now is None:
        now = arrow.utcnow()
    return (
        899
        - (
            now
            - now.replace(
                minute=(now.minute // 15) * 15, second=0, microsecond=0
            )
        ).seconds
    )


def get_seconds_until_next_day(now: arrow.arrow.Arrow | None = None) -> int:
    """
    Returns the number of seconds until the next day (utc midnight). This is the long-term rate limit used by Strava.
    :param now: A (utc) timestamp
    :type now: arrow.arrow.Arrow
    :return: the number of seconds until next day, as int
    """
    if now is None:
        now = arrow.utcnow()
    return (now.ceil("day") - now).seconds


class LimitStructure(TypedDict):
    """
    Dictionary that holds rate limits for a specific time window size.
    """

    usage: int
    """Usage in the current period"""

    limit: int
    """Limit for the period"""

    time: float
    """Duration of the period"""

    lastExceeded: datetime | None
    """Last time the limit was exceeded"""


class LimitsStructure(TypedDict):
    """
    Dictionary that holds rate limiting information for both of strava 15-minutes and
    daily rate limits.
    """

    short: LimitStructure
    """15-minutes rate limiting data"""

    long: LimitStructure
    """Daily rate limiting data"""


class XRateLimitRule:
    def __init__(
        self, limits: LimitsStructure, force_limits: bool = False
    ) -> None:
        """

        :param limits: The limits structure.
        :param force_limits:
            If False (default), this rule will set/update its limits based on
            what the Strava API tells it.
            If True, the provided limits will be enforced,
            i.e. ignoring the limits given by the API.
        """
        self.log = logging.getLogger(
            "{0.__module__}.{0.__name__}".format(self.__class__)
        )
        self.rate_limits = limits
        # should limit args be validated?
        self.limit_time_invalid: float = 0.0
        self.force_limits = force_limits

    @property
    def limit_timeout(self) -> float:
        return self.limit_time_invalid

    def __call__(self, response_headers: dict[str, str]) -> None:
        self._update_usage(response_headers)

        self._check_limit_time_invalid(self.rate_limits["short"])
        self._check_limit_rates(self.rate_limits["short"])
        self._check_limit_time_invalid(self.rate_limits["long"])
        self._check_limit_rates(self.rate_limits["long"])

    def _update_usage(self, response_headers: dict[str, str]) -> None:
        rates = get_rates_from_response_headers(response_headers)

        if rates:
            self.log.debug(
                "Updating rate-limit limits and usage from headers: {}".format(
                    rates
                )
            )
            self.rate_limits["short"]["usage"] = rates.short_usage
            self.rate_limits["long"]["usage"] = rates.long_usage

            if not self.force_limits:
                self.rate_limits["short"]["limit"] = rates.short_limit
                self.rate_limits["long"]["limit"] = rates.long_limit

    def _check_limit_rates(self, limit: LimitStructure) -> None:
        if limit["usage"] >= limit["limit"]:
            self.log.debug("Rate limit of {0} reached.".format(limit["limit"]))
            limit["lastExceeded"] = datetime.now()
            self._raise_rate_limit_exception(limit["limit"], limit["time"])

    def _check_limit_time_invalid(self, limit: LimitStructure) -> None:
        self.limit_time_invalid = 0
        if limit["lastExceeded"] is not None:
            delta = (datetime.now() - limit["lastExceeded"]).total_seconds()
            if delta < limit["time"]:
                self.limit_time_invalid = limit["time"] - delta
                self.log.debug(
                    "Rate limit invalid duration {0} seconds.".format(
                        self.limit_time_invalid
                    )
                )
                self._raise_rate_limit_timeout(
                    self.limit_timeout, limit["limit"]
                )

    def _raise_rate_limit_exception(
        self, timeout: float, limit_rate: float
    ) -> NoReturn:
        raise exc.RateLimitExceeded(
            "Rate limit of {0} exceeded. "
            "Try again in {1} seconds.".format(limit_rate, timeout),
            limit=limit_rate,
            timeout=timeout,
        )

    def _raise_rate_limit_timeout(
        self, timeout: float, limit_rate: float
    ) -> NoReturn:
        raise exc.RateLimitTimeout(
            "Rate limit of {} exceeded. "
            "Try again in {} seconds.".format(limit_rate, timeout),
            limit=limit_rate,
            timeout=timeout,
        )


class SleepingRateLimitRule:
    """
    A rate limit rule that can be prioritized and can dynamically adapt its limits based on API responses.
    Given its priority, it will enforce a variable "cool-down" period after each response. When rate limits
    are reached within their period, this limiter will wait until the end of that period. It will NOT raise
    any kind of exception in this case.
    """

    def __init__(
        self,
        priority: Literal["low", "medium", "high"] = "high",
        short_limit: int = 10000,
        long_limit: int = 1000000,
        force_limits: bool = False,
    ) -> None:
        """
        Constructs a new SleepingRateLimitRule.
        :param priority: The priority for this rule. When 'low', the cool-down period after each request will be such
        that the long-term limits will not be exceeded. When 'medium', the cool-down period will be such that the
        short-term limits will not be exceeded. When 'high', there will be no cool-down period.
        :type priority: str
        :param short_limit: (Optional) explicit short-term limit
        :type short_limit: int
        :param long_limit: (Optional) explicit long-term limit
        :type long_limit: int
        :param force_limits: If False (default), this rule will set/update its limits based on what the Strava API
        tells it. If True, the provided limits will be enforced, i.e. ignoring the limits given by the API.
        """
        if priority not in ["low", "medium", "high"]:
            raise ValueError(
                'Invalid priority "{0}", expecting one of "low", "medium" or "high"'.format(
                    priority
                )
            )

        self.log = logging.getLogger(
            "{0.__module__}.{0.__name__}".format(self.__class__)
        )
        self.priority = priority
        self.short_limit = short_limit
        self.long_limit = long_limit
        self.force_limits = force_limits

    def _get_wait_time(
        self,
        short_usage: int,
        long_usage: int,
        seconds_until_short_limit: int,
        seconds_until_long_limit: int,
    ) -> float:
        if long_usage >= self.long_limit:
            self.log.warning("Long term API rate limit exceeded")
            return seconds_until_long_limit
        elif short_usage >= self.short_limit:
            self.log.warning("Short term API rate limit exceeded")
            return seconds_until_short_limit

        if self.priority == "high":
            return 0
        elif self.priority == "medium":
            return seconds_until_short_limit / (self.short_limit - short_usage)
        elif self.priority == "low":
            return seconds_until_long_limit / (self.long_limit - long_usage)

    def __call__(self, response_headers: dict[str, str]) -> None:
        rates = get_rates_from_response_headers(response_headers)

        if rates:
            time.sleep(
                self._get_wait_time(
                    rates.short_usage,
                    rates.long_usage,
                    get_seconds_until_next_quarter(),
                    get_seconds_until_next_day(),
                )
            )
            if not self.force_limits:
                self.short_limit = rates.short_limit
                self.long_limit = rates.long_limit


class RateLimitRule:
    def __init__(
        self, requests: int, seconds: float, raise_exc: bool = False
    ) -> None:
        """
        :param requests: Number of requests for limit.
        :param seconds: The number of seconds for that number of requests (may be float)
        :param raise_exc: Whether to raise an exception when limit is reached (as opposed to pausing)
        """
        self.log = logging.getLogger(
            "{0.__module__}.{0.__name__}".format(self.__class__)
        )
        self.timeframe = timedelta(seconds=seconds)
        self.requests = requests
        self.tab: Deque[datetime] = collections.deque(maxlen=self.requests)
        self.raise_exc = raise_exc

    def __call__(self, args: dict[str, str]) -> None:
        """
        Register another request is being issued.

        Depending on configuration of the rule will pause if rate limit has
        been reached, or raise exception, etc.
        """
        # First check if the deque is full; that indicates that we'd better check whether
        # we need to pause.
        if len(self.tab) == self.requests:
            # Grab the oldest (leftmost) timestamp and check to see if it is greater than 1 second
            delta = datetime.now() - self.tab[0]
            if (
                delta < self.timeframe
            ):  # Has it been less than configured timeframe since oldest request?
                if self.raise_exc:
                    raise exc.RateLimitExceeded(
                        "Rate limit exceeded (can try again in {0})".format(
                            self.timeframe - delta
                        )
                    )
                else:
                    # Wait the difference between timeframe and the oldest request.
                    td = self.timeframe - delta
                    sleeptime = td.total_seconds()
                    self.log.debug(
                        "Rate limit triggered; sleeping for {0}".format(
                            sleeptime
                        )
                    )
                    time.sleep(sleeptime)
        self.tab.append(datetime.now())


class RateLimiter:
    def __init__(self) -> None:
        self.log: Logger = logging.getLogger(
            "{0.__module__}.{0.__name__}".format(self.__class__)
        )
        self.rules: list[Callable[[dict[str, str]], None]] = []

    def __call__(self, args: dict[str, str]) -> None:
        """
        Register another request is being issued.
        """
        for r in self.rules:
            r(args)


class DefaultRateLimiter(RateLimiter):
    """
    Implements something similar to the default rate limit for Strava apps.

    To do this correctly we would actually need to change our logic to reset
    the limit at midnight, etc.  Will make this more complex in the future.

    Strava API usage is limited on a per-application basis using a short term,
    15 minute, limit and a long term, daily, limit. The default rate limit allows
    600 requests every 15 minutes, with up to 30,000 requests per day.
    """

    def __init__(self) -> None:
        """
        Strava API usage is limited on a per-application basis using a short term,
        15 minute, limit and a long term, daily, limit. The default rate limit
        allows 600 requests every 15 minutes, with up to 30,000 requests per day.
        This limit allows applications to make 40 requests per minute for about half the day.
        """

        super().__init__()

        self.rules.append(
            XRateLimitRule(
                {
                    "short": {
                        "usage": 0,
                        # 60s * 15 = 15 min
                        "limit": 600,
                        "time": (60 * 15),
                        "lastExceeded": None,
                    },
                    "long": {
                        "usage": 0,
                        # 60s * 60m * 24 = 1 day
                        "limit": 30000,
                        "time": (60 * 60 * 24),
                        "lastExceeded": None,
                    },
                }
            )
        )

        # XRateLimitRule used instead of timer based RateLimitRule
        # self.rules.append(RateLimitRule(requests=40, seconds=60, raise_exc=False))
        # self.rules.append(RateLimitRule(requests=30000, seconds=(3600 * 24), raise_exc=True))



import functools

from stravalib import model
from stravalib.client import BatchedResultsIterator
from stravalib.tests.functional import FunctionalTestBase


class ResultIteratorTest(FunctionalTestBase):
    def setUp(self):
        super(ResultIteratorTest, self).setUp()
        self.protocol = self.client.protocol

    def test_limit_call(self):
        """Test setting the limit in method call."""

        result_fetcher = functools.partial(
            self.protocol.get, "/athlete/activities"
        )
        results = BatchedResultsIterator(
            entity=model.Activity,
            result_fetcher=result_fetcher,
            limit=10,
            per_page=2,
        )
        results = list(results)
        self.assertEqual(10, len(results))

    def test_multiple_iterator_calls(self):
        """Test multiple calls of the iterator."""

        result_fetcher = functools.partial(
            self.protocol.get, "/athlete/activities"
        )
        results = BatchedResultsIterator(
            entity=model.Activity,
            result_fetcher=result_fetcher,
            limit=10,
            per_page=2,
        )
        results.limit = 10
        results1 = list(results)
        results2 = list(results)

        self.assertEqual(10, len(results1))
        self.assertEqual(len(results1), len(results2))

    def test_limit_iterator(self):
        """Test setting the limit on the iterator."""

        result_fetcher = functools.partial(
            self.protocol.get, "/athlete/activities"
        )
        results = BatchedResultsIterator(
            entity=model.Activity,
            result_fetcher=result_fetcher,
            limit=10,
            per_page=2,
        )
        results.limit = 10
        results = list(results)
        self.assertEqual(10, len(results))

        # TODO: use a mock here to figure out how many calls are happening under the hood.

    def test_empty(self):
        """Test iterating over empty results."""

        # Specify two thing that we happen to know will return 0 results
        def pretend_fetcher(page, per_page):
            return []

        ri = BatchedResultsIterator(
            entity=model.Shoe, result_fetcher=pretend_fetcher
        )
        results = list(ri)
        self.assertEqual(0, len(results))



import os
from configparser import NoOptionError, SafeConfigParser

from stravalib.client import Client
from stravalib.tests import TESTS_DIR, TestBase

TEST_CFG = os.path.join(TESTS_DIR, "test.ini")


class FunctionalTestBase(TestBase):
    def setUp(self):
        super(FunctionalTestBase, self).setUp()
        if not os.path.exists(TEST_CFG):
            raise Exception(
                "Unable to run the write tests without a test.ini in that defines an access_token with write privs."
            )

        cfg = SafeConfigParser()
        with open(TEST_CFG) as fp:
            cfg.readfp(fp, "test.ini")
            access_token = cfg.get("write_tests", "access_token")
            try:
                activity_id = cfg.get("activity_tests", "activity_id")
            except NoOptionError:
                activity_id = None

        self.client = Client(access_token=access_token)
        self.activity_id = activity_id



from datetime import datetime, timedelta
from typing import List, Optional

import pint
import pytest
import pytz
from pydantic import BaseModel

from stravalib import model
from stravalib import unithelper as uh
from stravalib.model import (
    Activity,
    ActivityLap,
    ActivityPhoto,
    ActivityTotals,
    BackwardCompatibilityMixin,
    BaseEffort,
    BoundClientEntity,
    Club,
    LatLon,
    Segment,
    SegmentExplorerResult,
    SubscriptionCallback,
)
from stravalib.strava_model import LatLng
from stravalib.tests import TestBase
from stravalib.unithelper import Quantity, UnitConverter


@pytest.mark.parametrize("model_class,attr,value", ((Club, "name", "foo"),))
class TestLegacyModelSerialization:
    def test_legacy_deserialize(self, model_class, attr, value):
        with pytest.warns(DeprecationWarning):
            model_obj = model_class.deserialize({attr: value})
            assert getattr(model_obj, attr) == value

    def test_legacy_from_dict(self, model_class, attr, value):
        with pytest.warns(DeprecationWarning):
            model_obj = model_class()
            model_obj.from_dict({attr: value})
            assert getattr(model_obj, attr) == value

    def test_legacy_to_dict(self, model_class, attr, value):
        with pytest.warns(DeprecationWarning):
            model_obj = model_class(**{attr: value})
            model_dict_legacy = model_obj.to_dict()
            model_dict_modern = model_obj.dict()
            assert model_dict_legacy == model_dict_modern


@pytest.mark.parametrize(
    "model_class,raw,expected_value",
    (
        (Club, {"name": "foo"}, "foo"),
        (ActivityTotals, {"elapsed_time": 100}, timedelta(seconds=100)),
        (
            ActivityTotals,
            {"distance": 100.0},
            UnitConverter("meters")(100.0),
        ),
        (
            Activity,
            {"timezone": "Europe/Amsterdam"},
            pytz.timezone("Europe/Amsterdam"),
        ),
        (Club, {"activity_types": ["Run", "Ride"]}, ["Run", "Ride"]),
        (Activity, {"sport_type": "Run"}, "Run"),
    ),
)
def test_backward_compatibility_mixin_field_conversions(
    model_class, raw, expected_value
):
    obj = model_class.parse_obj(raw)
    assert getattr(obj, list(raw.keys())[0]) == expected_value


@pytest.mark.parametrize(
    "model_class,raw,expected_value",
    (
        (Activity, {"start_latlng": "5.4,4.3"}, LatLon(__root__=[5.4, 4.3])),
        (Activity, {"start_latlng": []}, None),
        (Segment, {"start_latlng": []}, None),
        (SegmentExplorerResult, {"start_latlng": []}, None),
        (ActivityPhoto, {"location": []}, None),
        (Activity, {"timezone": "foobar"}, None),
        (
            Activity,
            {"start_date_local": "2023-01-17T11:06:07Z"},
            datetime(2023, 1, 17, 11, 6, 7),
        ),
        (
            BaseEffort,
            {"start_date_local": "2023-01-17T11:06:07Z"},
            datetime(2023, 1, 17, 11, 6, 7),
        ),
        (
            ActivityLap,
            {"start_date_local": "2023-01-17T11:06:07Z"},
            datetime(2023, 1, 17, 11, 6, 7),
        ),
    ),
)
def test_deserialization_edge_cases(model_class, raw, expected_value):
    obj = model_class.parse_obj(raw)
    assert getattr(obj, list(raw.keys())[0]) == expected_value


def test_subscription_callback_field_names():
    sub_callback_raw = {
        "hub.mode": "subscribe",
        "hub.verify_token": "STRAVA",
        "hub.challenge": "15f7d1a91c1f40f8a748fd134752feb3",
    }
    sub_callback = SubscriptionCallback.parse_obj(sub_callback_raw)
    assert sub_callback.hub_mode == "subscribe"
    assert sub_callback.hub_verify_token == "STRAVA"


# Below are some toy classes to test type extensions and attribute lookup:
class A(BaseModel, BackwardCompatibilityMixin):
    x: Optional[int] = None


class ConversionA(A, BackwardCompatibilityMixin):
    _field_conversions = {"x": uh.meters}


class BoundA(A, BoundClientEntity):
    pass


class B(BaseModel, BackwardCompatibilityMixin):
    a: Optional[A] = None
    bound_a: Optional[BoundA] = None


class BoundB(B, BoundClientEntity):
    pass


class C(BaseModel, BackwardCompatibilityMixin):
    a: Optional[List[A]] = None
    bound_a: Optional[List[BoundA]] = None


class BoundC(C, BoundClientEntity):
    pass


class D(BaseModel, BackwardCompatibilityMixin):
    a: Optional[LatLng] = None


class ExtA(A):
    def foo(self):
        return self.x


class ExtLatLng(LatLng):
    def foo(self):
        return f"[{self[0], self[1]}]"


@pytest.mark.parametrize(
    "lookup_expression,expected_result,expected_bound_client",
    (
        (A().x, None, False),
        (B().a, None, False),
        (A(x=1).x, 1, False),
        (B(a=A(x=1)).a, A(x=1), False),
        (C(a=[A(x=1), A(x=2)]).a[1], A(x=2), False),
        (ConversionA(x=1).x, pint.Quantity("1 meter"), False),
        (BoundA(x=1).x, 1, False),
        (B(bound_a=BoundA(x=1)).bound_a, BoundA(x=1), None),
        (
            B(bound_a=BoundA(x=1, bound_client=1)).bound_a,
            BoundA(x=1, bound_client=1),
            True,
        ),
        (BoundB(a=A(x=1)).a, A(x=1), False),
        (BoundB(a=A(x=1), bound_client=1).a, A(x=1), False),
        (BoundB(bound_a=BoundA(x=1)).bound_a, BoundA(x=1), None),
        (
            BoundB(bound_a=BoundA(x=1), bound_client=1).bound_a,
            BoundA(x=1, bound_client=1),
            True,
        ),
        (C(bound_a=[BoundA(x=1), BoundA(x=2)]).bound_a[1], BoundA(x=2), None),
        (
            C(
                bound_a=[
                    BoundA(x=1, bound_client=1),
                    BoundA(x=2, bound_client=1),
                ]
            ).bound_a[1],
            BoundA(x=2, bound_client=1),
            True,
        ),
        (BoundC(a=[A(x=1), A(x=2)]).a[1], A(x=2), False),
        (BoundC(a=[A(x=1), A(x=2)], bound_client=1).a[1], A(x=2), False),
        (
            BoundC(bound_a=[BoundA(x=1), BoundA(x=2)]).bound_a[1],
            BoundA(x=2),
            None,
        ),
        (
            BoundC(bound_a=[BoundA(x=1), BoundA(x=2)], bound_client=1).bound_a[
                1
            ],
            BoundA(x=2, bound_client=1),
            True,
        ),
    ),
)
def test_backward_compatible_attribute_lookup(
    lookup_expression, expected_result, expected_bound_client
):
    assert lookup_expression == expected_result

    if expected_bound_client:
        assert lookup_expression.bound_client is not None
    elif expected_bound_client is None:
        assert lookup_expression.bound_client is None
    elif not expected_bound_client:
        assert not hasattr(lookup_expression, "bound_client")


class ModelTest(TestBase):
    def setUp(self):
        super(ModelTest, self).setUp()

    def test_entity_collections(self):
        a = model.Athlete()
        d = {
            "clubs": [
                {"resource_state": 2, "id": 7, "name": "Team Roaring Mouse"},
                {"resource_state": 2, "id": 1, "name": "Team Strava Cycling"},
                {
                    "resource_state": 2,
                    "id": 34444,
                    "name": "Team Strava Cyclocross",
                },
            ]
        }
        a.from_dict(d)

        self.assertEqual(3, len(a.clubs))
        self.assertEqual("Team Roaring Mouse", a.clubs[0].name)

    def test_speed_units(self):
        a = model.Activity()

        a.max_speed = 1000  # m/s
        a.average_speed = 1000  # m/s
        self.assertAlmostEqual(3600.0, float(uh.kph(a.max_speed)))
        self.assertAlmostEqual(3600.0, float(uh.kph(a.average_speed)))

        a.max_speed = uh.mph(1.0)
        # print repr(a.max_speed)

        self.assertAlmostEqual(1.61, float(uh.kph(a.max_speed)), places=2)

    def test_time_intervals(self):
        segment = model.Segment()
        # s.pr_time = XXXX

        split = model.Split()
        split.moving_time = 3.1
        split.elapsed_time = 5.73

    def test_distance_units(self):
        # Gear
        g = model.Gear()
        g.distance = 1000
        self.assertEqual(1.0, float(uh.kilometers(g.distance)))

        # Metric Split
        split = model.Split()
        split.distance = 1000  # meters
        split.elevation_difference = 1000  # meters
        self.assertIsInstance(split.distance, Quantity)
        self.assertIsInstance(split.elevation_difference, Quantity)
        self.assertEqual(1.0, float(uh.kilometers(split.distance)))
        self.assertEqual(1.0, float(uh.kilometers(split.elevation_difference)))
        split = None

        # Segment
        s = model.Segment()
        s.distance = 1000
        s.elevation_high = 2000
        s.elevation_low = 1000
        self.assertIsInstance(s.distance, Quantity)
        self.assertIsInstance(s.elevation_high, Quantity)
        self.assertIsInstance(s.elevation_low, Quantity)
        self.assertEqual(1.0, float(uh.kilometers(s.distance)))
        self.assertEqual(2.0, float(uh.kilometers(s.elevation_high)))
        self.assertEqual(1.0, float(uh.kilometers(s.elevation_low)))

        # Activity
        a = model.Activity()
        a.distance = 1000  # m
        a.total_elevation_gain = 1000  # m
        self.assertIsInstance(a.distance, Quantity)
        self.assertIsInstance(a.total_elevation_gain, Quantity)
        self.assertEqual(1.0, float(uh.kilometers(a.distance)))
        self.assertEqual(1.0, float(uh.kilometers(a.total_elevation_gain)))

    def test_weight_units(self):
        """ """
        # PowerActivityZone

    def test_subscription_deser(self):
        d = {
            "id": 1,
            "object_type": "activity",
            "aspect_type": "create",
            "callback_url": "http://you.com/callback/",
            "created_at": "2015-04-29T18:11:09.400558047-07:00",
            "updated_at": "2015-04-29T18:11:09.400558047-07:00",
        }
        sub = model.Subscription.parse_obj(d)
        self.assertEqual(d["id"], sub.id)

    def test_subscription_update_deser(self):
        d = {
            "subscription_id": "1",
            "owner_id": 13408,
            "object_id": 12312312312,
            "object_type": "activity",
            "aspect_type": "create",
            "event_time": 1297286541,
        }
        subupd = model.SubscriptionUpdate.deserialize(d)
        self.assertEqual(
            "2011-02-09 21:22:21",
            subupd.event_time.strftime("%Y-%m-%d %H:%M:%S"),
        )



#!/usr/bin/env python
"""
A basic authorization server.  Run this with your Strava Client ID and Client Secret and access from your
browser (because the Strava OAuth page uses javascript) in order to get a resulting access token.  That access
token can then be used to initialize a Client that can read (and/or write) data from the Strava API.

You must run this from a virtualenv that has stravalib installed.

Example Usage:

  (env) shell$ python -m stravalib.tests.auth_responder --port=8000 --client-id=123 --client-secret=deadbeefdeadbeefdeadbeefdeadbeefdeadbeef

  Then connect in your browser to http://localhost:8000/

  The redirected response (from Strava) will deliver a code that can be exchanged for a token.  The access token will be
  presented in the browser after the exchange.  Save this value into your config (e.g. into your test.ini) to run
  functional tests.
"""
import argparse
import logging
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib import parse as urlparse

from stravalib import Client


class StravaAuthHTTPServer(HTTPServer):
    def __init__(
        self,
        server_address,
        RequestHandlerClass,
        client_id,
        client_secret,
        bind_and_activate=True,
    ):
        HTTPServer.__init__(
            self,
            server_address,
            RequestHandlerClass,
            bind_and_activate=bind_and_activate,
        )
        self.logger = logging.getLogger("auth_server.http")
        self.client_id = client_id
        self.client_secret = client_secret
        self.listening_event = threading.Event()

    def serve_forever(self, *args, **kwargs):
        self.listening_event.set()
        return HTTPServer.serve_forever(self, *args, **kwargs)


class RequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        request_path = self.path

        parsed_path = urlparse.urlparse(request_path)

        client = Client()

        if request_path.startswith("/authorization"):
            self.send_response(200)
            self.send_header(b"Content-type", b"text/plain")
            self.end_headers()

            self.wfile.write(b"Authorization Handler\n\n")
            code = urlparse.parse_qs(parsed_path.query).get("code")
            if code:
                code = code[0]
                token_response = client.exchange_code_for_token(
                    client_id=self.server.client_id,
                    client_secret=self.server.client_secret,
                    code=code,
                )
                access_token = token_response["access_token"]
                self.server.logger.info(
                    "Exchanged code {} for access token {}".format(
                        code, access_token
                    )
                )
                self.wfile.write(b"Access Token: {}\n".format(access_token))
            else:
                self.server.logger.error("No code param received.")
                self.wfile.write(b"ERROR: No code param recevied.\n")
        else:
            url = client.authorization_url(
                client_id=self.server.client_id,
                redirect_uri="http://localhost:{}/authorization".format(
                    self.server.server_port
                ),
            )

            self.send_response(302)
            self.send_header("Content-type", "text/plain")
            self.send_header(b"Location", bytes(url))
            self.end_headers()
            self.wfile.write(b"Redirect to URL: {}\n".format(url))


def main(port, client_id, client_secret):
    logging.basicConfig(
        level=logging.INFO, format="%(levelname)-8s %(message)s"
    )

    logger = logging.getLogger("auth_responder")
    logger.info("Listening on localhost:%s" % port)

    server = StravaAuthHTTPServer(
        ("", port),
        RequestHandler,
        client_id=client_id,
        client_secret=client_secret,
    )
    server.serve_forever()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Run a local web server to receive authorization responses from Strava."
    )

    parser.add_argument(
        "-p",
        "--port",
        help="Which port to bind to",
        action="store",
        type=int,
        default=8000,
    )

    parser.add_argument(
        "--client-id",
        help="Strava API Client ID",
        action="store",
        required=True,
    )
    parser.add_argument(
        "--client-secret",
        help="Strava API Client Secret",
        action="store",
        required=True,
    )
    args = parser.parse_args()

    main(
        port=args.port,
        client_id=args.client_id,
        client_secret=args.client_secret,
    )



import time

from stravalib import exc
from stravalib.tests.functional import FunctionalTestBase
from stravalib.util.limiter import DefaultRateLimiter, XRateLimitRule


class ClientDefaultRateLimiterTest(FunctionalTestBase):
    def test_fail_on_rate_limit_exceeded(self):
        """Use this test as an example"""

        # setup 'short' limit for testing
        self.client.protocol.rate_limiter.rules = []
        self.client.protocol.rate_limiter.rules.append(
            XRateLimitRule(
                {
                    "short": {
                        "usage": 0,
                        "limit": 600,
                        "time": 5,
                        "lastExceeded": None,
                    },
                    "long": {
                        "usage": 0,
                        "limit": 30000,
                        "time": 5,
                        "lastExceeded": None,
                    },
                }
            )
        )

        # interact with api to get the limits
        self.client.get_athlete()

        # acces the default rate limit rule
        rate_limit_rule = self.client.protocol.rate_limiter.rules[0]

        # get any of the rate limits, ex the 'short'
        limit = rate_limit_rule.rate_limits["short"]

        # get current usage
        usage = limit["usage"]
        print("last rate limit usage is {0}".format(usage))

        # for testing purpses set the limit to usage
        limit["limit"] = usage
        print("changing limit to {0}".format(limit["limit"]))

        # expect exception because of RateLimit has been
        #  exceeded (or reached max)
        with self.assertRaises(exc.RateLimitExceeded):
            self.client.get_athlete()

        # request fired to early (less than 5 sec) causes timeout exception
        with self.assertRaises(exc.RateLimitTimeout):
            self.client.get_athlete()

        # once rate limit has exceeded wait until another reuqest is possible
        #  check if timout has been set
        self.assertTrue(rate_limit_rule.limit_timeout > 0)
        print("limit timeout {0}".format(rate_limit_rule.limit_timeout))

        # reseting limit
        # simulates Strava api - it would set the usage again to 0
        limit["limit"] = 600
        print("resetting limit to {0}".format(limit["limit"]))

        try:
            # waiting until timeout expires
            time.sleep(5)

            # this time it should work again
            self.client.get_athlete()
            self.assertTrue("No exception raised")
        except exc.RateLimitExceeded as e:
            self.fail("limiter raised RateLimitTimeout unexpectedly!")

        # continuse other tests with DefaultRateLimiter
        print("setting default rate limiter")
        self.client.protocol.rate_limiter = DefaultRateLimiter()



import pytest
import pytz

from stravalib.field_conversions import (
    enum_value,
    enum_values,
    optional_input,
    timezone,
)
from stravalib.strava_model import ActivityType, SportType


def test_optional_input():
    @optional_input
    def foo(x):
        return x + 1

    assert foo(1) == 2
    assert foo(None) is None


def test_enum_value():
    assert enum_value(ActivityType(__root__="Run")) == "Run"


def test_enum_values():
    assert enum_values(
        [ActivityType(__root__="Run"), SportType(__root__="Ride")]
    ) == ["Run", "Ride"]


@pytest.mark.parametrize(
    "arg,expected_value",
    (
        ("Factory", None),
        ("(GMT+00:00) Factory", None),
        ("Europe/Amsterdam", pytz.timezone("Europe/Amsterdam")),
        ("(GMT+01:00) Europe/Amsterdam", pytz.timezone("Europe/Amsterdam")),
    ),
)
def test_timezone(arg, expected_value):
    assert timezone(arg) == expected_value



import datetime

import requests

from stravalib import model
from stravalib import unithelper as uh
from stravalib.tests.functional import FunctionalTestBase


class ClientTest(FunctionalTestBase):
    def test_get_starred_segment(self):
        """
        Test get_starred_segment
        """
        i = 0
        for segment in self.client.get_starred_segment(limit=5):
            self.assertIsInstance(segment, model.Segment)
            i += 1
        self.assertGreater(i, 0)  # star at least one segment
        self.assertLessEqual(i, 5)

    def test_get_activity(self):
        """Test basic activity fetching."""
        activity = self.client.get_activity(96089609)
        self.assertEqual("El Dorado County, CA, USA", activity.location_city)

        self.assertIsInstance(activity.start_latlng, model.LatLon)
        self.assertAlmostEqual(
            -120.4357631, activity.start_latlng.lon, places=2
        )
        self.assertAlmostEqual(
            38.74263759999999, activity.start_latlng.lat, places=2
        )

        self.assertIsInstance(activity.map, model.Map)

        self.assertIsInstance(activity.athlete, model.Athlete)
        self.assertEqual(1513, activity.athlete.id)

        # self.assertAlmostEqual(first, second, places, msg, delta)
        # Ensure that iw as read in with correct units
        self.assertEqual(22.5308, float(uh.kilometers(activity.distance)))

    def test_get_activity_and_segments(self):
        """Test include_all_efforts parameter on activity fetching."""
        if not self.activity_id:
            self.fail(
                "Include an activity_id in test.ini to test segment_efforts"
            )

        activity = self.client.get_activity(
            self.activity_id, include_all_efforts=True
        )
        self.assertTrue(isinstance(activity.segment_efforts, list))

        # Check also when we have no parameters segment_efforts is None
        activity_no_segments = self.client.get_activity(self.activity_id)
        self.assertTrue(activity.segment_efforts, None)

    def test_get_route(self):
        route = self.client.get_route(3445913)

        self.assertEqual("Baveno - Mottarone", route.name)
        self.assertAlmostEqual(
            1265.20, float(uh.meters(route.elevation_gain)), 2
        )

    def test_get_activity_laps(self):
        activity = self.client.get_activity(165094211)
        laps = list(self.client.get_activity_laps(165094211))
        self.assertEqual(5, len(laps))
        # This obviously is far from comprehensive, just a sanity check
        self.assertEqual("Lap 1", laps[0].name)
        self.assertEqual(178.0, laps[0].max_heartrate)

    def test_get_activity_zones(self):
        """
        Test loading zones for activity.
        """
        zones = self.client.get_activity_zones(99895560)
        print(zones)
        self.assertEqual(1, len(zones))
        self.assertIsInstance(zones[0], model.PaceActivityZone)

        # Indirectly
        activity = self.client.get_activity(99895560)
        self.assertEqual(len(zones), len(activity.zones))
        self.assertEqual(zones[0].score, activity.zones[0].score)

    def test_activity_comments(self):
        """
        Test loading comments for already-loaded activity.
        """
        activity = self.client.get_activity(2290897)
        self.assertTrue(activity.comment_count > 0)
        comments = list(activity.comments)
        self.assertEqual(3, len(comments))
        self.assertEqual(
            "I love Gordo's. I've been eating there for 20 years!",
            comments[0].text,
        )

    def test_activity_photos(self):
        """
        Test photos on activity
        """
        activity = self.client.get_activity(643026323)
        self.assertTrue(activity.total_photo_count > 0)
        photos = list(activity.full_photos)
        self.assertEqual(len(photos), 1)
        self.assertEqual(len(photos), activity.total_photo_count)
        self.assertIsInstance(photos[0], model.ActivityPhoto)

    def test_activity_kudos(self):
        """
        Test kudos on activity
        """
        activity = self.client.get_activity(152668627)
        self.assertTrue(activity.kudos_count > 0)
        kudos = list(activity.kudos)
        self.assertGreater(len(kudos), 6)
        self.assertEqual(len(kudos), activity.kudos_count)
        self.assertIsInstance(kudos[0], model.ActivityKudos)

    def test_activity_streams(self):
        """
        Test activity streams
        """
        stypes = [
            "time",
            "latlng",
            "distance",
            "altitude",
            "velocity_smooth",
            "heartrate",
            "cadence",
            "watts",
            "temp",
            "moving",
            "grade_smooth",
        ]

        streams = self.client.get_activity_streams(152668627, stypes, "low")

        self.assertGreater(len(streams.keys()), 3)
        for k in streams.keys():
            self.assertIn(k, stypes)

        # time stream
        self.assertIsInstance(streams["time"].data[0], int)
        self.assertGreater(streams["time"].original_size, 100)
        self.assertEqual(streams["time"].resolution, "low")
        self.assertEqual(len(streams["time"].data), 100)

        # latlng stream
        self.assertIsInstance(streams["latlng"].data, list)
        self.assertIsInstance(streams["latlng"].data[0][0], float)

    def test_route_streams(self):
        """
        Test toute streams
        """
        stypes = ["latlng", "distance", "altitude"]

        streams = self.client.get_route_streams(3445913)
        self.assertEqual(len(streams.keys()), 3)

        for t in stypes:
            self.assertIn(t, streams.keys())

    def test_related_activities(self):
        """
        Test get_related_activities on an activity and related property of Activity
        """
        activity_id = 152668627
        activity = self.client.get_activity(activity_id)
        related_activities = list(
            self.client.get_related_activities(activity_id)
        )

        # Check the number of related_activities matches what activity would expect
        self.assertEqual(len(related_activities), activity.athlete_count - 1)

        # Check the related property gives the same result
        related_activities_from_property = list(activity.related)
        self.assertEqual(related_activities, related_activities_from_property)

    def test_effort_streams(self):
        """
        Test effort streams
        """
        stypes = ["distance"]

        activity = self.client.get_activity(165479860)  # 152668627)
        streams = self.client.get_effort_streams(
            activity.segment_efforts[0].id, stypes, "medium"
        )

        self.assertIn("distance", streams.keys())

        # distance stream
        self.assertIsInstance(streams["distance"].data[0], float)  # xxx
        self.assertEqual(streams["distance"].resolution, "medium")
        self.assertEqual(
            len(streams["distance"].data),
            min(1000, streams["distance"].original_size),
        )

    def test_get_curr_athlete(self):
        athlete = self.client.get_athlete()

        # Just some basic sanity checks here
        self.assertTrue(len(athlete.firstname) > 0)

        self.assertTrue(athlete.athlete_type in ["runner", "cyclist"])

    def test_get_athlete_clubs(self):
        clubs = self.client.get_athlete_clubs()
        self.assertEqual(3, len(clubs))
        self.assertEqual("Team Roaring Mouse", clubs[0].name)
        self.assertEqual("Team Strava Cycling", clubs[1].name)
        self.assertEqual("Team Strava Cyclocross", clubs[2].name)

        clubs_indirect = self.client.get_athlete().clubs
        self.assertEqual(3, len(clubs_indirect))
        self.assertEqual(clubs[0].name, clubs_indirect[0].name)
        self.assertEqual(clubs[1].name, clubs_indirect[1].name)
        self.assertEqual(clubs[2].name, clubs_indirect[2].name)

    def test_get_gear(self):
        g = self.client.get_gear("g69911")
        self.assertTrue(float(g.distance) >= 3264.67)
        self.assertEqual("Salomon XT Wings 2", g.name)
        self.assertEqual("Salomon", g.brand_name)
        self.assertTrue(g.primary)
        self.assertEqual(model.DETAILED, g.resource_state)
        self.assertEqual("g69911", g.id)
        self.assertEqual("XT Wings 2", g.model_name)
        self.assertEqual("", g.description)

    def test_get_segment(self):
        segment = self.client.get_segment(229781)
        self.assertIsInstance(segment, model.Segment)
        print(segment)
        self.assertEqual("Hawk Hill", segment.name)
        self.assertAlmostEqual(
            2.68, float(uh.kilometers(segment.distance)), places=2
        )

        # Fetch leaderboard
        lb = segment.leaderboard
        self.assertEqual(10, len(lb))  # 10 top results, 5 bottom results

    def test_get_segment_efforts(self):
        # test with string
        efforts = self.client.get_segment_efforts(
            4357415,
            start_date_local="2012-12-23T00:00:00Z",
            end_date_local="2012-12-23T11:00:00Z",
        )
        print(efforts)

        i = 0
        for effort in efforts:
            print(effort)
            self.assertEqual(4357415, effort.segment.id)
            self.assertIsInstance(effort, model.BaseEffort)
            effort_date = effort.start_date_local
            self.assertEqual(effort_date.strftime("%Y-%m-%d"), "2012-12-23")
            i += 1
        print(i)

        self.assertGreater(i, 2)

        # also test with datetime object
        start_date = datetime.datetime(2012, 12, 31, 6, 0)
        end_date = start_date + datetime.timedelta(hours=12)
        efforts = self.client.get_segment_efforts(
            4357415,
            start_date_local=start_date,
            end_date_local=end_date,
        )
        print(efforts)

        i = 0
        for effort in efforts:
            print(effort)
            self.assertEqual(4357415, effort.segment.id)
            self.assertIsInstance(effort, model.BaseEffort)
            effort_date = effort.start_date_local
            self.assertEqual(effort_date.strftime("%Y-%m-%d"), "2012-12-31")
            i += 1
        print(i)

        self.assertGreater(i, 2)

    def test_segment_explorer(self):
        bounds = (37.821362, -122.505373, 37.842038, -122.465977)
        results = self.client.explore_segments(bounds)

        # This might be brittle
        self.assertEqual("Hawk Hill", results[0].name)

        # Fetch full segment
        segment = results[0].segment
        self.assertEqual(results[0].name, segment.name)

        # For some reason these don't follow the simple math rules one might expect (so we round to int)
        self.assertAlmostEqual(
            results[0].elev_difference,
            segment.elevation_high - segment.elevation_low,
            places=0,
        )


class AthleteStatsTest(FunctionalTestBase):
    """
    Tests the functionality for collecting athlete statistics
    https://developers.strava.com/docs/reference/#api-Athletes-getStats
    """

    def test_basic_get_from_client(self):
        stats = self.client.get_athlete_stats()
        self.assertIsInstance(stats, model.AthleteStats)
        self.assertIsInstance(stats.recent_ride_totals, model.ActivityTotals)
        print("Biggest climb: {!r}".format(stats.biggest_climb_elevation_gain))

        # Check biggest_climb_elevation_gain has been set
        self.assertTrue(
            uh.meters(stats.biggest_climb_elevation_gain) >= uh.meters(0)
        )

    def test_get_from_client_with_authenticated_id(self):
        athlete_id = self.client.get_athlete().id
        stats = self.client.get_athlete_stats(athlete_id)
        self.assertIsInstance(stats, model.AthleteStats)
        # Check same as before
        self.assertEqual(
            stats.biggest_climb_elevation_gain,
            self.client.get_athlete_stats().biggest_climb_elevation_gain,
        )

    def test_get_from_client_with_wrong_id(self):
        CAV_ID = 1353775
        # Currently raises a requests.exceptions.HTTPError, TODO: better error handling
        self.assertRaises(
            requests.exceptions.HTTPError,
            self.client.get_athlete_stats,
            CAV_ID,
        )

    def test_athlete_stats_property_option(self):
        a = self.client.get_athlete()
        stats = a.stats
        self.assertIsInstance(stats, model.AthleteStats)

    def test_athlete_stats_cached(self):
        a = self.client.get_athlete()
        a._stats = "Not None"
        stats = a.stats
        self.assertEqual(stats, "Not None")

    def test_athlete_property_not_authenticated(self):
        with self.assertRaises(NotImplementedError):
            cav = self.client.get_athlete(1353775)



import arrow

from stravalib.tests import TestBase
from stravalib.util.limiter import (
    SleepingRateLimitRule,
    XRateLimitRule,
    get_rates_from_response_headers,
    get_seconds_until_next_day,
    get_seconds_until_next_quarter,
)

test_response = {
    "Status": "404 Not Found",
    "X-Request-Id": "a1a4a4973962ffa7e0f18d7c485fe741",
    "Content-Encoding": "gzip",
    "Content-Length": "104",
    "Connection": "keep-alive",
    "X-RateLimit-Limit": "600,30000",
    "X-UA-Compatible": "IE=Edge,chrome=1",
    "Cache-Control": "no-cache, private",
    "Date": "Tue, 14 Nov 2017 11:29:15 GMT",
    "X-FRAME-OPTIONS": "DENY",
    "Content-Type": "application/json; charset=UTF-8",
    "X-RateLimit-Usage": "4,67",
}

test_response_no_rates = {
    "Status": "200 OK",
    "X-Request-Id": "d465159561420f6e0239dc24429a7cf3",
    "Content-Encoding": "gzip",
    "Content-Length": "371",
    "Connection": "keep-alive",
    "X-UA-Compatible": "IE=Edge,chrome=1",
    "Cache-Control": "max-age=0, private, must-revalidate",
    "Date": "Tue, 14 Nov 2017 13:19:31 GMT",
    "X-FRAME-OPTIONS": "DENY",
    "Content-Type": "application/json; charset=UTF-8",
}


class LimiterTest(TestBase):
    def test_get_rates_from_response_headers(self):
        """Should return namedtuple with rates"""
        request_rates = get_rates_from_response_headers(test_response)
        self.assertEqual(600, request_rates.short_limit)
        self.assertEqual(30000, request_rates.long_limit)
        self.assertEqual(4, request_rates.short_usage)
        self.assertEqual(67, request_rates.long_usage)

    def test_get_rates_from_response_headers_missing_rates(self):
        """Should return namedtuple with None values for rates in case of missing rates in headers"""
        self.assertIsNone(
            get_rates_from_response_headers(test_response_no_rates)
        )

    def test_get_seconds_until_next_quarter(self):
        """Should return number of seconds to next quarter of an hour"""
        self.assertEqual(
            59,
            get_seconds_until_next_quarter(
                arrow.get(2017, 11, 1, 17, 14, 0, 0)
            ),
        )
        self.assertEqual(
            59,
            get_seconds_until_next_quarter(
                arrow.get(2017, 11, 1, 17, 59, 0, 0)
            ),
        )
        self.assertEqual(
            0,
            get_seconds_until_next_quarter(
                arrow.get(2017, 11, 1, 17, 59, 59, 999999)
            ),
        )
        self.assertEqual(
            899,
            get_seconds_until_next_quarter(
                arrow.get(2017, 11, 1, 17, 0, 0, 1)
            ),
        )

    def test_get_seconds_until_next_day(self):
        """Should return the number of seconds until next day"""
        self.assertEqual(
            59,
            get_seconds_until_next_day(arrow.get(2017, 11, 1, 23, 59, 0, 0)),
        )
        self.assertEqual(
            86399,
            get_seconds_until_next_day(arrow.get(2017, 11, 1, 0, 0, 0, 0)),
        )


class XRateLimitRuleTest(TestBase):
    def test_rule_normal_response(self):
        rule = XRateLimitRule(
            {
                "short": {
                    "usage": 0,
                    "limit": 600,
                    "time": (60 * 15),
                    "lastExceeded": None,
                },
                "long": {
                    "usage": 0,
                    "limit": 30000,
                    "time": (60 * 60 * 24),
                    "lastExceeded": None,
                },
            }
        )
        rule(test_response)
        self.assertEqual(4, rule.rate_limits["short"]["usage"])
        self.assertEqual(67, rule.rate_limits["long"]["usage"])

    def test_rule_missing_rates_response(self):
        rule = XRateLimitRule(
            {
                "short": {
                    "usage": 0,
                    "limit": 600,
                    "time": (60 * 15),
                    "lastExceeded": None,
                },
                "long": {
                    "usage": 0,
                    "limit": 30000,
                    "time": (60 * 60 * 24),
                    "lastExceeded": None,
                },
            }
        )
        rule(test_response_no_rates)
        self.assertEqual(0, rule.rate_limits["short"]["usage"])
        self.assertEqual(0, rule.rate_limits["long"]["usage"])


class SleepingRateLimitRuleTest(TestBase):
    def setUp(self):
        self.test_response = test_response.copy()

    def test_invalid_priority(self):
        """Should raise ValueError in case of invalid priority"""
        with self.assertRaises(ValueError):
            SleepingRateLimitRule(priority="foobar")

    def test_get_wait_time_high_priority(self):
        """Should never sleep/wait after high priority requests"""
        self.assertEqual(
            0, SleepingRateLimitRule()._get_wait_time(42, 42, 60, 3600)
        )

    def test_get_wait_time_medium_priority(self):
        """Should return number of seconds to next short limit divided by number of remaining requests
        for that period"""
        rule = SleepingRateLimitRule(priority="medium", short_limit=11)
        self.assertEqual(1, rule._get_wait_time(1, 1, 10, 1000))
        self.assertEqual(0.5, rule._get_wait_time(1, 1, 5, 1000))

    def test_get_wait_time_low_priority(self):
        """Should return number of seconds to next long limit divided by number of remaining requests
        for that period"""
        rule = SleepingRateLimitRule(priority="low", long_limit=11)
        self.assertEqual(1, rule._get_wait_time(1, 1, 1, 10))
        self.assertEqual(0.5, rule._get_wait_time(1, 1, 1, 5))

    def test_get_wait_time_limit_reached(self):
        """Should wait until end of period when limit is reached, regardless priority"""
        rule = SleepingRateLimitRule(short_limit=10, long_limit=100)
        self.assertEqual(42, rule._get_wait_time(10, 10, 42, 1000))
        self.assertEqual(42, rule._get_wait_time(1234, 10, 42, 1000))
        self.assertEqual(21, rule._get_wait_time(10, 100, 42, 21))
        self.assertEqual(21, rule._get_wait_time(10, 1234, 42, 21))

    def test_invocation_unchanged_limits(self):
        """Should not update limits if these don't change"""
        self.test_response["X-RateLimit-Usage"] = "0, 0"
        self.test_response["X-RateLimit-Limit"] = "10000, 1000000"
        rule = SleepingRateLimitRule()
        self.assertEqual(10000, rule.short_limit)
        self.assertEqual(1000000, rule.long_limit)
        rule(self.test_response)
        self.assertEqual(10000, rule.short_limit)
        self.assertEqual(1000000, rule.long_limit)

    def test_invocation_changed_limits(self):
        """Should update limits in case of changes, depending on limit enforcement"""
        self.test_response["X-RateLimit-Usage"] = "0, 0"
        self.test_response["X-RateLimit-Limit"] = "600, 30000"

        # without limit enforcement (default)
        rule = SleepingRateLimitRule()
        rule(self.test_response)
        self.assertEqual(600, rule.short_limit)
        self.assertEqual(30000, rule.long_limit)

        # with limit enforcement
        rule = SleepingRateLimitRule(force_limits=True)
        rule(self.test_response)
        self.assertEqual(10000, rule.short_limit)
        self.assertEqual(1000000, rule.long_limit)



#!flask/bin/python
import logging

from flask import Flask, jsonify, redirect, render_template, request, url_for

from stravalib import Client

app = Flask(__name__)
app.config.from_envvar("APP_SETTINGS")


@app.route("/")
def login():
    c = Client()
    url = c.authorization_url(
        client_id=app.config["STRAVA_CLIENT_ID"],
        redirect_uri=url_for(".logged_in", _external=True),
        approval_prompt="auto",
    )
    return render_template("login.html", authorize_url=url)


@app.route("/strava-oauth")
def logged_in():
    """
    Method called by Strava (redirect) that includes parameters.
    - state
    - code
    - error
    """
    error = request.args.get("error")
    state = request.args.get("state")
    if error:
        return render_template("login_error.html", error=error)
    else:
        code = request.args.get("code")
        client = Client()
        access_token = client.exchange_code_for_token(
            client_id=app.config["STRAVA_CLIENT_ID"],
            client_secret=app.config["STRAVA_CLIENT_SECRET"],
            code=code,
        )
        # Probably here you'd want to store this somewhere -- e.g. in a database.
        strava_athlete = client.get_athlete()

        return render_template(
            "login_results.html",
            athlete=strava_athlete,
            access_token=access_token,
        )


if __name__ == "__main__":
    app.run(debug=True)



# generated by datamodel-codegen:
#   filename:  <stdin>

from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Field, conint


class ActivityTotal(BaseModel):
    """
    A roll-up of metrics pertaining to a set of activities. Values are in seconds and meters.
    """

    achievement_count: Optional[int] = None
    """
    The total number of achievements of the considered activities.
    """
    count: Optional[int] = None
    """
    The number of activities considered in this total.
    """
    distance: Optional[float] = None
    """
    The total distance covered by the considered activities.
    """
    elapsed_time: Optional[int] = None
    """
    The total elapsed time of the considered activities.
    """
    elevation_gain: Optional[float] = None
    """
    The total elevation gain of the considered activities.
    """
    moving_time: Optional[int] = None
    """
    The total moving time of the considered activities.
    """


class ActivityType(BaseModel):
    __root__: Literal[
        "AlpineSki",
        "BackcountrySki",
        "Canoeing",
        "Crossfit",
        "EBikeRide",
        "Elliptical",
        "Golf",
        "Handcycle",
        "Hike",
        "IceSkate",
        "InlineSkate",
        "Kayaking",
        "Kitesurf",
        "NordicSki",
        "Ride",
        "RockClimbing",
        "RollerSki",
        "Rowing",
        "Run",
        "Sail",
        "Skateboard",
        "Snowboard",
        "Snowshoe",
        "Soccer",
        "StairStepper",
        "StandUpPaddling",
        "Surfing",
        "Swim",
        "Velomobile",
        "VirtualRide",
        "VirtualRun",
        "Walk",
        "WeightTraining",
        "Wheelchair",
        "Windsurf",
        "Workout",
        "Yoga",
    ]
    """
    An enumeration of the types an activity may have. Note that this enumeration does not include new sport types (e.g. MountainBikeRide, EMountainBikeRide), activities with these sport types will have the corresponding activity type (e.g. Ride for MountainBikeRide, EBikeRide for EMountainBikeRide)
    """


class BaseStream(BaseModel):
    original_size: Optional[int] = None
    """
    The number of data points in this stream
    """
    resolution: Optional[Literal["low", "medium", "high"]] = None
    """
    The level of detail (sampling) in which this stream was returned
    """
    series_type: Optional[Literal["distance", "time"]] = None
    """
    The base series used in the case the stream was downsampled
    """


class CadenceStream(BaseStream):
    data: Optional[List[int]] = None
    """
    The sequence of cadence values for this stream, in rotations per minute
    """


class ClubAthlete(BaseModel):
    admin: Optional[bool] = None
    """
    Whether the athlete is a club admin.
    """
    firstname: Optional[str] = None
    """
    The athlete's first name.
    """
    lastname: Optional[str] = None
    """
    The athlete's last initial.
    """
    member: Optional[str] = None
    """
    The athlete's member status.
    """
    owner: Optional[bool] = None
    """
    Whether the athlete is club owner.
    """
    resource_state: Optional[int] = None
    """
    Resource state, indicates level of detail. Possible values: 1 -> "meta", 2 -> "summary", 3 -> "detail"
    """


class DistanceStream(BaseStream):
    data: Optional[List[float]] = None
    """
    The sequence of distance values for this stream, in meters
    """


class Error(BaseModel):
    code: Optional[str] = None
    """
    The code associated with this error.
    """
    field: Optional[str] = None
    """
    The specific field or aspect of the resource associated with this error.
    """
    resource: Optional[str] = None
    """
    The type of resource associated with this error.
    """


class Fault(BaseModel):
    """
    Encapsulates the errors that may be returned from the API.
    """

    errors: Optional[List[Error]] = None
    """
    The set of specific errors associated with this fault, if any.
    """
    message: Optional[str] = None
    """
    The message of the fault.
    """


class HeartrateStream(BaseStream):
    data: Optional[List[int]] = None
    """
    The sequence of heart rate values for this stream, in beats per minute
    """


class LatLng(BaseModel):
    """
    A pair of latitude/longitude coordinates, represented as an array of 2 floating point numbers.
    """

    __root__: List[float] = Field(..., max_items=2, min_items=2)
    """
    A pair of latitude/longitude coordinates, represented as an array of 2 floating point numbers.
    """


class LatLngStream(BaseStream):
    data: Optional[List[LatLng]] = None
    """
    The sequence of lat/long values for this stream
    """


class MembershipApplication(BaseModel):
    active: Optional[bool] = None
    """
    Whether the membership is currently active
    """
    membership: Optional[Literal["member", "pending"]] = None
    """
    The membership status of this application
    """
    success: Optional[bool] = None
    """
    Whether the application for membership was successfully submitted
    """


class MetaActivity(BaseModel):
    id: Optional[int] = None
    """
    The unique identifier of the activity
    """


class MetaAthlete(BaseModel):
    id: Optional[int] = None
    """
    The unique identifier of the athlete
    """


class MetaClub(BaseModel):
    id: Optional[int] = None
    """
    The club's unique identifier.
    """
    name: Optional[str] = None
    """
    The club's name.
    """
    resource_state: Optional[int] = None
    """
    Resource state, indicates level of detail. Possible values: 1 -> "meta", 2 -> "summary", 3 -> "detail"
    """


class MovingStream(BaseStream):
    data: Optional[List[bool]] = None
    """
    The sequence of moving values for this stream, as boolean values
    """


class Primary(BaseModel):
    id: Optional[int] = None
    source: Optional[int] = None
    unique_id: Optional[str] = None
    urls: Optional[Dict[str, str]] = None


class PhotosSummary(BaseModel):
    count: Optional[int] = None
    """
    The number of photos
    """
    primary: Optional[Primary] = None


class PolylineMap(BaseModel):
    id: Optional[str] = None
    """
    The identifier of the map
    """
    polyline: Optional[str] = None
    """
    The polyline of the map, only returned on detailed representation of an object
    """
    summary_polyline: Optional[str] = None
    """
    The summary polyline of the map
    """


class PowerStream(BaseStream):
    data: Optional[List[int]] = None
    """
    The sequence of power values for this stream, in watts
    """


class SmoothGradeStream(BaseStream):
    data: Optional[List[float]] = None
    """
    The sequence of grade values for this stream, as percents of a grade
    """


class SmoothVelocityStream(BaseStream):
    data: Optional[List[float]] = None
    """
    The sequence of velocity values for this stream, in meters per second
    """


class Split(BaseModel):
    average_speed: Optional[float] = None
    """
    The average speed of this split, in meters per second
    """
    distance: Optional[float] = None
    """
    The distance of this split, in meters
    """
    elapsed_time: Optional[int] = None
    """
    The elapsed time of this split, in seconds
    """
    elevation_difference: Optional[float] = None
    """
    The elevation difference of this split, in meters
    """
    moving_time: Optional[int] = None
    """
    The moving time of this split, in seconds
    """
    pace_zone: Optional[int] = None
    """
    The pacing zone of this split
    """
    split: Optional[int] = None
    """
    N/A
    """


class SportType(BaseModel):
    __root__: Literal[
        "AlpineSki",
        "BackcountrySki",
        "Badminton",
        "Canoeing",
        "Crossfit",
        "EBikeRide",
        "Elliptical",
        "EMountainBikeRide",
        "Golf",
        "GravelRide",
        "Handcycle",
        "HighIntensityIntervalTraining",
        "Hike",
        "IceSkate",
        "InlineSkate",
        "Kayaking",
        "Kitesurf",
        "MountainBikeRide",
        "NordicSki",
        "Pickleball",
        "Pilates",
        "Racquetball",
        "Ride",
        "RockClimbing",
        "RollerSki",
        "Rowing",
        "Run",
        "Sail",
        "Skateboard",
        "Snowboard",
        "Snowshoe",
        "Soccer",
        "Squash",
        "StairStepper",
        "StandUpPaddling",
        "Surfing",
        "Swim",
        "TableTennis",
        "Tennis",
        "TrailRun",
        "Velomobile",
        "VirtualRide",
        "VirtualRow",
        "VirtualRun",
        "Walk",
        "WeightTraining",
        "Wheelchair",
        "Windsurf",
        "Workout",
        "Yoga",
    ]
    """
    An enumeration of the sport types an activity may have. Distinct from ActivityType in that it has new types (e.g. MountainBikeRide)
    """


class StreamType(BaseModel):
    __root__: Literal[
        "time",
        "distance",
        "latlng",
        "altitude",
        "velocity_smooth",
        "heartrate",
        "cadence",
        "watts",
        "temp",
        "moving",
        "grade_smooth",
    ]
    """
    An enumeration of the supported types of streams.
    """


class SummaryActivity(MetaActivity):
    achievement_count: Optional[int] = None
    """
    The number of achievements gained during this activity
    """
    athlete: Optional[MetaAthlete] = None
    athlete_count: Optional[conint(ge=1)] = None
    """
    The number of athletes for taking part in a group activity
    """
    average_speed: Optional[float] = None
    """
    The activity's average speed, in meters per second
    """
    average_watts: Optional[float] = None
    """
    Average power output in watts during this activity. Rides only
    """
    comment_count: Optional[int] = None
    """
    The number of comments for this activity
    """
    commute: Optional[bool] = None
    """
    Whether this activity is a commute
    """
    device_watts: Optional[bool] = None
    """
    Whether the watts are from a power meter, false if estimated
    """
    distance: Optional[float] = None
    """
    The activity's distance, in meters
    """
    elapsed_time: Optional[int] = None
    """
    The activity's elapsed time, in seconds
    """
    elev_high: Optional[float] = None
    """
    The activity's highest elevation, in meters
    """
    elev_low: Optional[float] = None
    """
    The activity's lowest elevation, in meters
    """
    end_latlng: Optional[LatLng] = None
    external_id: Optional[str] = None
    """
    The identifier provided at upload time
    """
    flagged: Optional[bool] = None
    """
    Whether this activity is flagged
    """
    gear_id: Optional[str] = None
    """
    The id of the gear for the activity
    """
    has_kudoed: Optional[bool] = None
    """
    Whether the logged-in athlete has kudoed this activity
    """
    hide_from_home: Optional[bool] = None
    """
    Whether the activity is muted
    """
    kilojoules: Optional[float] = None
    """
    The total work done in kilojoules during this activity. Rides only
    """
    kudos_count: Optional[int] = None
    """
    The number of kudos given for this activity
    """
    manual: Optional[bool] = None
    """
    Whether this activity was created manually
    """
    map: Optional[PolylineMap] = None
    max_speed: Optional[float] = None
    """
    The activity's max speed, in meters per second
    """
    max_watts: Optional[int] = None
    """
    Rides with power meter data only
    """
    moving_time: Optional[int] = None
    """
    The activity's moving time, in seconds
    """
    name: Optional[str] = None
    """
    The name of the activity
    """
    photo_count: Optional[int] = None
    """
    The number of Instagram photos for this activity
    """
    private: Optional[bool] = None
    """
    Whether this activity is private
    """
    sport_type: Optional[SportType] = None
    start_date: Optional[datetime] = None
    """
    The time at which the activity was started.
    """
    start_date_local: Optional[datetime] = None
    """
    The time at which the activity was started in the local timezone.
    """
    start_latlng: Optional[LatLng] = None
    timezone: Optional[str] = None
    """
    The timezone of the activity
    """
    total_elevation_gain: Optional[float] = None
    """
    The activity's total elevation gain.
    """
    total_photo_count: Optional[int] = None
    """
    The number of Instagram and Strava photos for this activity
    """
    trainer: Optional[bool] = None
    """
    Whether this activity was recorded on a training machine
    """
    type: Optional[ActivityType] = None
    """
    Deprecated. Prefer to use sport_type
    """
    upload_id: Optional[int] = None
    """
    The identifier of the upload that resulted in this activity
    """
    upload_id_str: Optional[str] = None
    """
    The unique identifier of the upload in string format
    """
    weighted_average_watts: Optional[int] = None
    """
    Similar to Normalized Power. Rides with power meter data only
    """
    workout_type: Optional[int] = None
    """
    The activity's workout type
    """


class SummaryAthlete(MetaAthlete):
    city: Optional[str] = None
    """
    The athlete's city.
    """
    country: Optional[str] = None
    """
    The athlete's country.
    """
    created_at: Optional[datetime] = None
    """
    The time at which the athlete was created.
    """
    firstname: Optional[str] = None
    """
    The athlete's first name.
    """
    lastname: Optional[str] = None
    """
    The athlete's last name.
    """
    premium: Optional[bool] = None
    """
    Deprecated.  Use summit field instead. Whether the athlete has any Summit subscription.
    """
    profile: Optional[str] = None
    """
    URL to a 124x124 pixel profile picture.
    """
    profile_medium: Optional[str] = None
    """
    URL to a 62x62 pixel profile picture.
    """
    resource_state: Optional[int] = None
    """
    Resource state, indicates level of detail. Possible values: 1 -> "meta", 2 -> "summary", 3 -> "detail"
    """
    sex: Optional[Literal["M", "F"]] = None
    """
    The athlete's sex.
    """
    state: Optional[str] = None
    """
    The athlete's state or geographical region.
    """
    summit: Optional[bool] = None
    """
    Whether the athlete has any Summit subscription.
    """
    updated_at: Optional[datetime] = None
    """
    The time at which the athlete was last updated.
    """


class SummaryClub(MetaClub):
    activity_types: Optional[List[ActivityType]] = None
    """
    The activity types that count for a club. This takes precedence over sport_type.
    """
    city: Optional[str] = None
    """
    The club's city.
    """
    country: Optional[str] = None
    """
    The club's country.
    """
    cover_photo: Optional[str] = None
    """
    URL to a ~1185x580 pixel cover photo.
    """
    cover_photo_small: Optional[str] = None
    """
    URL to a ~360x176  pixel cover photo.
    """
    featured: Optional[bool] = None
    """
    Whether the club is featured or not.
    """
    member_count: Optional[int] = None
    """
    The club's member count.
    """
    private: Optional[bool] = None
    """
    Whether the club is private.
    """
    profile_medium: Optional[str] = None
    """
    URL to a 60x60 pixel profile picture.
    """
    sport_type: Optional[
        Literal["cycling", "running", "triathlon", "other"]
    ] = None
    """
    Deprecated. Prefer to use activity_types.
    """
    state: Optional[str] = None
    """
    The club's state or geographical region.
    """
    url: Optional[str] = None
    """
    The club's vanity URL.
    """
    verified: Optional[bool] = None
    """
    Whether the club is verified or not.
    """


class SummaryGear(BaseModel):
    distance: Optional[float] = None
    """
    The distance logged with this gear.
    """
    id: Optional[str] = None
    """
    The gear's unique identifier.
    """
    name: Optional[str] = None
    """
    The gear's name.
    """
    primary: Optional[bool] = None
    """
    Whether this gear's is the owner's default one.
    """
    resource_state: Optional[int] = None
    """
    Resource state, indicates level of detail. Possible values: 2 -> "summary", 3 -> "detail"
    """


class SummaryPRSegmentEffort(BaseModel):
    effort_count: Optional[int] = None
    """
    Number of efforts by the authenticated athlete on this segment.
    """
    pr_activity_id: Optional[int] = None
    """
    The unique identifier of the activity related to the PR effort.
    """
    pr_date: Optional[datetime] = None
    """
    The time at which the PR effort was started.
    """
    pr_elapsed_time: Optional[int] = None
    """
    The elapsed time ot the PR effort.
    """


class SummarySegmentEffort(BaseModel):
    activity_id: Optional[int] = None
    """
    The unique identifier of the activity related to this effort
    """
    distance: Optional[float] = None
    """
    The effort's distance in meters
    """
    elapsed_time: Optional[int] = None
    """
    The effort's elapsed time
    """
    id: Optional[int] = None
    """
    The unique identifier of this effort
    """
    is_kom: Optional[bool] = None
    """
    Whether this effort is the current best on the leaderboard
    """
    start_date: Optional[datetime] = None
    """
    The time at which the effort was started.
    """
    start_date_local: Optional[datetime] = None
    """
    The time at which the effort was started in the local timezone.
    """


class TemperatureStream(BaseStream):
    data: Optional[List[int]] = None
    """
    The sequence of temperature values for this stream, in celsius degrees
    """


class TimeStream(BaseStream):
    data: Optional[List[int]] = None
    """
    The sequence of time values for this stream, in seconds
    """


class UpdatableActivity(BaseModel):
    commute: Optional[bool] = None
    """
    Whether this activity is a commute
    """
    description: Optional[str] = None
    """
    The description of the activity
    """
    gear_id: Optional[str] = None
    """
    Identifier for the gear associated with the activity. ‘none’ clears gear from activity
    """
    hide_from_home: Optional[bool] = None
    """
    Whether this activity is muted
    """
    name: Optional[str] = None
    """
    The name of the activity
    """
    sport_type: Optional[SportType] = None
    trainer: Optional[bool] = None
    """
    Whether this activity was recorded on a training machine
    """
    type: Optional[ActivityType] = None
    """
    Deprecated. Prefer to use sport_type. In a request where both type and sport_type are present, this field will be ignored
    """


class Upload(BaseModel):
    activity_id: Optional[int] = None
    """
    The identifier of the activity this upload resulted into
    """
    error: Optional[str] = None
    """
    The error associated with this upload
    """
    external_id: Optional[str] = None
    """
    The external identifier of the upload
    """
    id: Optional[int] = None
    """
    The unique identifier of the upload
    """
    id_str: Optional[str] = None
    """
    The unique identifier of the upload in string format
    """
    status: Optional[str] = None
    """
    The status of this upload
    """


class ZoneRange(BaseModel):
    max: Optional[int] = None
    """
    The maximum value in the range.
    """
    min: Optional[int] = None
    """
    The minimum value in the range.
    """


class ZoneRanges(BaseModel):
    __root__: List[ZoneRange]


class ActivityStats(BaseModel):
    """
    A set of rolled-up statistics and totals for an athlete
    """

    all_ride_totals: Optional[ActivityTotal] = None
    """
    The all time ride stats for the athlete.
    """
    all_run_totals: Optional[ActivityTotal] = None
    """
    The all time run stats for the athlete.
    """
    all_swim_totals: Optional[ActivityTotal] = None
    """
    The all time swim stats for the athlete.
    """
    biggest_climb_elevation_gain: Optional[float] = None
    """
    The highest climb ridden by the athlete.
    """
    biggest_ride_distance: Optional[float] = None
    """
    The longest distance ridden by the athlete.
    """
    recent_ride_totals: Optional[ActivityTotal] = None
    """
    The recent (last 4 weeks) ride stats for the athlete.
    """
    recent_run_totals: Optional[ActivityTotal] = None
    """
    The recent (last 4 weeks) run stats for the athlete.
    """
    recent_swim_totals: Optional[ActivityTotal] = None
    """
    The recent (last 4 weeks) swim stats for the athlete.
    """
    ytd_ride_totals: Optional[ActivityTotal] = None
    """
    The year to date ride stats for the athlete.
    """
    ytd_run_totals: Optional[ActivityTotal] = None
    """
    The year to date run stats for the athlete.
    """
    ytd_swim_totals: Optional[ActivityTotal] = None
    """
    The year to date swim stats for the athlete.
    """


class AltitudeStream(BaseStream):
    data: Optional[List[float]] = None
    """
    The sequence of altitude values for this stream, in meters
    """


class ClubActivity(BaseModel):
    athlete: Optional[MetaAthlete] = None
    distance: Optional[float] = None
    """
    The activity's distance, in meters
    """
    elapsed_time: Optional[int] = None
    """
    The activity's elapsed time, in seconds
    """
    moving_time: Optional[int] = None
    """
    The activity's moving time, in seconds
    """
    name: Optional[str] = None
    """
    The name of the activity
    """
    sport_type: Optional[SportType] = None
    total_elevation_gain: Optional[float] = None
    """
    The activity's total elevation gain.
    """
    type: Optional[ActivityType] = None
    """
    Deprecated. Prefer to use sport_type
    """
    workout_type: Optional[int] = None
    """
    The activity's workout type
    """


class ClubAnnouncement(BaseModel):
    athlete: Optional[SummaryAthlete] = None
    club_id: Optional[int] = None
    """
    The unique identifier of the club this announcements was made in.
    """
    created_at: Optional[datetime] = None
    """
    The time at which this announcement was created.
    """
    id: Optional[int] = None
    """
    The unique identifier of this announcement.
    """
    message: Optional[str] = None
    """
    The content of this announcement
    """


class Comment(BaseModel):
    activity_id: Optional[int] = None
    """
    The identifier of the activity this comment is related to
    """
    athlete: Optional[SummaryAthlete] = None
    created_at: Optional[datetime] = None
    """
    The time at which this comment was created.
    """
    id: Optional[int] = None
    """
    The unique identifier of this comment
    """
    text: Optional[str] = None
    """
    The content of the comment
    """


class DetailedAthlete(SummaryAthlete):
    bikes: Optional[List[SummaryGear]] = None
    """
    The athlete's bikes.
    """
    clubs: Optional[List[SummaryClub]] = None
    """
    The athlete's clubs.
    """
    follower_count: Optional[int] = None
    """
    The athlete's follower count.
    """
    friend_count: Optional[int] = None
    """
    The athlete's friend count.
    """
    ftp: Optional[int] = None
    """
    The athlete's FTP (Functional Threshold Power).
    """
    measurement_preference: Optional[Literal["feet", "meters"]] = None
    """
    The athlete's preferred unit system.
    """
    shoes: Optional[List[SummaryGear]] = None
    """
    The athlete's shoes.
    """
    weight: Optional[float] = None
    """
    The athlete's weight.
    """


class DetailedClub(SummaryClub):
    admin: Optional[bool] = None
    """
    Whether the currently logged-in athlete is an administrator of this club.
    """
    following_count: Optional[int] = None
    """
    The number of athletes in the club that the logged-in athlete follows.
    """
    membership: Optional[Literal["member", "pending"]] = None
    """
    The membership status of the logged-in athlete.
    """
    owner: Optional[bool] = None
    """
    Whether the currently logged-in athlete is the owner of this club.
    """


class DetailedGear(SummaryGear):
    brand_name: Optional[str] = None
    """
    The gear's brand name.
    """
    description: Optional[str] = None
    """
    The gear's description.
    """
    frame_type: Optional[int] = None
    """
    The gear's frame type (bike only).
    """
    model_name: Optional[str] = None
    """
    The gear's model name.
    """


class ExplorerSegment(BaseModel):
    avg_grade: Optional[float] = None
    """
    The segment's average grade, in percents
    """
    climb_category: Optional[conint(ge=0, le=5)] = None
    """
    The category of the climb [0, 5]. Higher is harder ie. 5 is Hors catégorie, 0 is uncategorized in climb_category. If climb_category = 5, climb_category_desc = HC. If climb_category = 2, climb_category_desc = 3.
    """
    climb_category_desc: Optional[
        Literal["NC", "4", "3", "2", "1", "HC"]
    ] = None
    """
    The description for the category of the climb
    """
    distance: Optional[float] = None
    """
    The segment's distance, in meters
    """
    elev_difference: Optional[float] = None
    """
    The segments's evelation difference, in meters
    """
    end_latlng: Optional[LatLng] = None
    id: Optional[int] = None
    """
    The unique identifier of this segment
    """
    name: Optional[str] = None
    """
    The name of this segment
    """
    points: Optional[str] = None
    """
    The polyline of the segment
    """
    start_latlng: Optional[LatLng] = None


class HeartRateZoneRanges(BaseModel):
    custom_zones: Optional[bool] = None
    """
    Whether the athlete has set their own custom heart rate zones
    """
    zones: Optional[ZoneRanges] = None


class Lap(BaseModel):
    activity: Optional[MetaActivity] = None
    athlete: Optional[MetaAthlete] = None
    average_cadence: Optional[float] = None
    """
    The lap's average cadence
    """
    average_speed: Optional[float] = None
    """
    The lap's average speed
    """
    distance: Optional[float] = None
    """
    The lap's distance, in meters
    """
    elapsed_time: Optional[int] = None
    """
    The lap's elapsed time, in seconds
    """
    end_index: Optional[int] = None
    """
    The end index of this effort in its activity's stream
    """
    id: Optional[int] = None
    """
    The unique identifier of this lap
    """
    lap_index: Optional[int] = None
    """
    The index of this lap in the activity it belongs to
    """
    max_speed: Optional[float] = None
    """
    The maximum speed of this lat, in meters per second
    """
    moving_time: Optional[int] = None
    """
    The lap's moving time, in seconds
    """
    name: Optional[str] = None
    """
    The name of the lap
    """
    pace_zone: Optional[int] = None
    """
    The athlete's pace zone during this lap
    """
    split: Optional[int] = None
    start_date: Optional[datetime] = None
    """
    The time at which the lap was started.
    """
    start_date_local: Optional[datetime] = None
    """
    The time at which the lap was started in the local timezone.
    """
    start_index: Optional[int] = None
    """
    The start index of this effort in its activity's stream
    """
    total_elevation_gain: Optional[float] = None
    """
    The elevation gain of this lap, in meters
    """


class PowerZoneRanges(BaseModel):
    zones: Optional[ZoneRanges] = None


class StreamSet(BaseModel):
    altitude: Optional[AltitudeStream] = None
    cadence: Optional[CadenceStream] = None
    distance: Optional[DistanceStream] = None
    grade_smooth: Optional[SmoothGradeStream] = None
    heartrate: Optional[HeartrateStream] = None
    latlng: Optional[LatLngStream] = None
    moving: Optional[MovingStream] = None
    temp: Optional[TemperatureStream] = None
    time: Optional[TimeStream] = None
    velocity_smooth: Optional[SmoothVelocityStream] = None
    watts: Optional[PowerStream] = None


class SummarySegment(BaseModel):
    activity_type: Optional[Literal["Ride", "Run"]] = None
    athlete_pr_effort: Optional[SummaryPRSegmentEffort] = None
    athlete_segment_stats: Optional[SummarySegmentEffort] = None
    average_grade: Optional[float] = None
    """
    The segment's average grade, in percents
    """
    city: Optional[str] = None
    """
    The segments's city.
    """
    climb_category: Optional[int] = None
    """
    The category of the climb [0, 5]. Higher is harder ie. 5 is Hors catégorie, 0 is uncategorized in climb_category.
    """
    country: Optional[str] = None
    """
    The segment's country.
    """
    distance: Optional[float] = None
    """
    The segment's distance, in meters
    """
    elevation_high: Optional[float] = None
    """
    The segments's highest elevation, in meters
    """
    elevation_low: Optional[float] = None
    """
    The segments's lowest elevation, in meters
    """
    end_latlng: Optional[LatLng] = None
    id: Optional[int] = None
    """
    The unique identifier of this segment
    """
    maximum_grade: Optional[float] = None
    """
    The segments's maximum grade, in percents
    """
    name: Optional[str] = None
    """
    The name of this segment
    """
    private: Optional[bool] = None
    """
    Whether this segment is private.
    """
    start_latlng: Optional[LatLng] = None
    state: Optional[str] = None
    """
    The segments's state or geographical region.
    """


class TimedZoneRange(ZoneRange):
    """
    A union type representing the time spent in a given zone.
    """

    time: Optional[int] = None
    """
    The number of seconds spent in this zone
    """


class Zones(BaseModel):
    heart_rate: Optional[HeartRateZoneRanges] = None
    power: Optional[PowerZoneRanges] = None


class DetailedSegment(SummarySegment):
    athlete_count: Optional[int] = None
    """
    The number of unique athletes who have an effort for this segment
    """
    created_at: Optional[datetime] = None
    """
    The time at which the segment was created.
    """
    effort_count: Optional[int] = None
    """
    The total number of efforts for this segment
    """
    hazardous: Optional[bool] = None
    """
    Whether this segment is considered hazardous
    """
    map: Optional[PolylineMap] = None
    star_count: Optional[int] = None
    """
    The number of stars for this segment
    """
    total_elevation_gain: Optional[float] = None
    """
    The segment's total elevation gain.
    """
    updated_at: Optional[datetime] = None
    """
    The time at which the segment was last updated.
    """


class DetailedSegmentEffort(SummarySegmentEffort):
    activity: Optional[MetaActivity] = None
    athlete: Optional[MetaAthlete] = None
    average_cadence: Optional[float] = None
    """
    The effort's average cadence
    """
    average_heartrate: Optional[float] = None
    """
    The heart heart rate of the athlete during this effort
    """
    average_watts: Optional[float] = None
    """
    The average wattage of this effort
    """
    device_watts: Optional[bool] = None
    """
    For riding efforts, whether the wattage was reported by a dedicated recording device
    """
    end_index: Optional[int] = None
    """
    The end index of this effort in its activity's stream
    """
    hidden: Optional[bool] = None
    """
    Whether this effort should be hidden when viewed within an activity
    """
    kom_rank: Optional[conint(ge=1, le=10)] = None
    """
    The rank of the effort on the global leaderboard if it belongs in the top 10 at the time of upload
    """
    max_heartrate: Optional[float] = None
    """
    The maximum heart rate of the athlete during this effort
    """
    moving_time: Optional[int] = None
    """
    The effort's moving time
    """
    name: Optional[str] = None
    """
    The name of the segment on which this effort was performed
    """
    pr_rank: Optional[conint(ge=1, le=3)] = None
    """
    The rank of the effort on the athlete's leaderboard if it belongs in the top 3 at the time of upload
    """
    segment: Optional[SummarySegment] = None
    start_index: Optional[int] = None
    """
    The start index of this effort in its activity's stream
    """


class ExplorerResponse(BaseModel):
    segments: Optional[List[ExplorerSegment]] = None
    """
    The set of segments matching an explorer request
    """


class Route(BaseModel):
    athlete: Optional[SummaryAthlete] = None
    created_at: Optional[datetime] = None
    """
    The time at which the route was created
    """
    description: Optional[str] = None
    """
    The description of the route
    """
    distance: Optional[float] = None
    """
    The route's distance, in meters
    """
    elevation_gain: Optional[float] = None
    """
    The route's elevation gain.
    """
    estimated_moving_time: Optional[int] = None
    """
    Estimated time in seconds for the authenticated athlete to complete route
    """
    id: Optional[int] = None
    """
    The unique identifier of this route
    """
    id_str: Optional[str] = None
    """
    The unique identifier of the route in string format
    """
    map: Optional[PolylineMap] = None
    name: Optional[str] = None
    """
    The name of this route
    """
    private: Optional[bool] = None
    """
    Whether this route is private
    """
    segments: Optional[List[SummarySegment]] = None
    """
    The segments traversed by this route
    """
    starred: Optional[bool] = None
    """
    Whether this route is starred by the logged-in athlete
    """
    sub_type: Optional[int] = None
    """
    This route's sub-type (1 for road, 2 for mountain bike, 3 for cross, 4 for trail, 5 for mixed)
    """
    timestamp: Optional[int] = None
    """
    An epoch timestamp of when the route was created
    """
    type: Optional[int] = None
    """
    This route's type (1 for ride, 2 for runs)
    """
    updated_at: Optional[datetime] = None
    """
    The time at which the route was last updated
    """


class TimedZoneDistribution(BaseModel):
    """
    Stores the exclusive ranges representing zones and the time spent in each.
    """

    __root__: List[TimedZoneRange]
    """
    Stores the exclusive ranges representing zones and the time spent in each.
    """


class ActivityZone(BaseModel):
    custom_zones: Optional[bool] = None
    distribution_buckets: Optional[TimedZoneDistribution] = None
    max: Optional[int] = None
    points: Optional[int] = None
    score: Optional[int] = None
    sensor_based: Optional[bool] = None
    type: Optional[Literal["heartrate", "power"]] = None


class DetailedActivity(SummaryActivity):
    best_efforts: Optional[List[DetailedSegmentEffort]] = None
    calories: Optional[float] = None
    """
    The number of kilocalories consumed during this activity
    """
    description: Optional[str] = None
    """
    The description of the activity
    """
    device_name: Optional[str] = None
    """
    The name of the device used to record the activity
    """
    embed_token: Optional[str] = None
    """
    The token used to embed a Strava activity
    """
    gear: Optional[SummaryGear] = None
    laps: Optional[List[Lap]] = None
    photos: Optional[PhotosSummary] = None
    segment_efforts: Optional[List[DetailedSegmentEffort]] = None
    splits_metric: Optional[List[Split]] = None
    """
    The splits of this activity in metric units (for runs)
    """
    splits_standard: Optional[List[Split]] = None
    """
    The splits of this activity in imperial units (for runs)
    """



import os
from datetime import datetime, timedelta

from stravalib import exc, model
from stravalib import unithelper as uh
from stravalib.tests import RESOURCES_DIR
from stravalib.tests.functional import FunctionalTestBase


class ClientWriteTest(FunctionalTestBase):
    def test_create_activity(self):
        """
        Test Client.create_activity simple case.
        """
        now = datetime.now().replace(microsecond=0)
        a = self.client.create_activity(
            "test_create_activity#simple",
            activity_type="Ride",
            start_date_local=now,
            elapsed_time=timedelta(hours=3, minutes=4, seconds=5),
            distance=uh.miles(15.2),
        )
        print(a)

        self.assertIsInstance(a, model.Activity)
        self.assertEqual("test_create_activity#simple", a.name)
        self.assertEqual(now, a.start_date_local)
        self.assertEqual(
            round(float(uh.miles(15.2)), 2),
            round(float(uh.miles(a.distance)), 2),
        )
        self.assertEqual(
            timedelta(hours=3, minutes=4, seconds=5), a.elapsed_time
        )

    def test_update_activity(self):
        """
        Test Client.update_activity simple case.
        """
        now = datetime.now().replace(microsecond=0)
        a = self.client.create_activity(
            "test_update_activity#create",
            activity_type="Ride",
            start_date_local=now,
            elapsed_time=timedelta(hours=3, minutes=4, seconds=5),
            distance=uh.miles(15.2),
        )

        self.assertIsInstance(a, model.Activity)
        self.assertEqual("test_update_activity#create", a.name)

        update1 = self.client.update_activity(
            a.id, name="test_update_activivty#update"
        )
        self.assertEqual("test_update_activivty#update", update1.name)
        self.assertFalse(update1.private)
        self.assertFalse(update1.trainer)
        self.assertFalse(update1.commute)

        update2 = self.client.update_activity(a.id, private=True)
        self.assertTrue(update2.private)

        update3 = self.client.update_activity(a.id, trainer=True)
        self.assertTrue(update3.private)
        self.assertTrue(update3.trainer)

    def test_upload_activity(self):
        """
        Test uploading an activity.

        NOTE: This requires clearing out the uploaded activities from configured
        writable Strava acct.
        """
        with open(os.path.join(RESOURCES_DIR, "sample.tcx")) as fp:
            uploader = self.client.upload_activity(fp, data_type="tcx")
            self.assertTrue(uploader.is_processing)
            a = uploader.wait()
            self.assertTrue(uploader.is_complete)
            self.assertIsInstance(a, model.Activity)
            self.assertEqual("02/21/2009 Leiden, ZH, The Netherlands", a.name)

            # And we'll get an error if we try the same file again
            with self.assertRaises(exc.ActivityUploadFailed):
                self.client.upload_activity(fp, data_type="tcx")



import json
import logging
import os
import re
from functools import lru_cache, wraps
from typing import Any, Callable, Dict, Optional

import requests
from responses import BaseResponse, RequestsMock

from stravalib.protocol import ApiV3
from stravalib.tests import RESOURCES_DIR

LOGGER = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def _get_strava_api_paths():
    use_local = False

    try:
        strava_api_swagger_response = requests.get(
            "https://developers.strava.com/swagger/swagger.json"
        )
        if strava_api_swagger_response.status_code != 200:
            use_local = True
    except requests.exceptions.ConnectionError:
        use_local = True

    if use_local:
        LOGGER.warning(
            "Failed to retrieve recent swagger API definition from Strava, using "
            "(potentially stale) version from local resources"
        )
        with open(
            os.path.join(RESOURCES_DIR, "strava_swagger.json"), "r"
        ) as swagger_file:
            return json.load(swagger_file)["paths"]
    else:
        return strava_api_swagger_response.json()["paths"]


def _api_method_adapter(api_method: Callable) -> Callable:
    """
    Decorator for mock registration methods of `responses.RequestsMock`
    """

    @wraps(api_method)
    def method_wrapper(
        *args,
        response_update: Dict[str, Any] = None,
        n_results: Optional[int] = None,
        **kwargs,
    ) -> BaseResponse:
        """
        Wrapper for mock registration methods of `responses.RequestsMock`

        Parameters
        ----------
        response_update:
            Dict that will be used to update any JSON object defined as
            an example response in the API's swagger.json
        n_results:
            The number of example objects to be returned by the mock
            response (in case of an array response).
        *args:
            Url and/or other positional arguments that would otherwise
            be provided to the `responses` registration methods. Note
            that url is expected to be relative, matching one of the
            paths defined in the Strava API's swagger.json.
        **kwargs
            Keyword arguments that would otherwise be provided to the
            `responses` registration methods.

        Returns
        -------
        A mocked response registration object

        This wrapper expects a relative url that matches one of the
        paths in the swagger.json of Strava's API, e.g.,
        `/activities/{Id}`. For the given HTTP method and status code,
        the corresponding example JSON response is retrieved from the
        swagger.json, and used in the mock registration as the response
        body. This response can be updated with the optional
        `response_update` argument (e.g. to set specific fields). If
        the response is a JSON array, the `n_results` argument
        specifies how many objects are in the response.

        Alternatively, a complete response can be specified using the
        `json` or `body` keywords, as is usually done using the
        `responses` library. In that case, the response updating -
        and multiplication mechanisms above are ignored and the
        given response is returned as-is.
        """

        # get url from args/kwargs
        try:
            relative_url = args[0]
        except IndexError:
            try:
                relative_url = kwargs.pop("url")
            except KeyError:
                raise ValueError(
                    "Expecting url either as first positional argument or keyword argument"
                )
        # match url with swagger path
        path_info = _get_strava_api_paths()[relative_url]
        # find default response in swagger if no json response is provided in the kwargs
        if "json" not in kwargs:
            http_method = api_method.args[0].lower()
            response_status = kwargs.get("status", 200)

            try:
                method_responses = path_info[http_method]
            except KeyError:
                raise ValueError(
                    f"Endpoint {relative_url} has no support for method {http_method}"
                )

            try:
                response = method_responses["responses"][str(response_status)][
                    "examples"
                ]["application/json"]
                if isinstance(response, list) and n_results is not None:
                    # Make sure response has n_results items
                    response = (response * (n_results // len(response) + 1))[
                        :n_results
                    ]
                elif n_results is not None:
                    # Force single response in example into result list (not all examples provide lists)
                    LOGGER.warning(
                        f"Forcing example single response into list"
                    )
                    response = [{**response, **response_update}] * n_results
            except KeyError:
                LOGGER.warning(
                    f"There are no known example responses for HTTP status {response_status}, "
                    f"using empty response. You may want to provide a full json response "
                    f'using the "json" keyword argument.'
                )
                response = {}

            # update fields if necessary
            if response_update is not None:
                if isinstance(response, list):
                    response = [
                        {**item, **response_update} for item in response
                    ]
                else:
                    response.update(response_update)

            kwargs.update({"json": response})

        # replace named parameters in url by wildcards
        matching_url = re.sub(r"\{\w+\}", r"\\w+", relative_url)
        return api_method(
            re.compile(
                ApiV3().resolve_url(matching_url)
            ),  # replaces url from args[0]
            *args[1:],
            **kwargs,
        )

    return method_wrapper


class StravaAPIMock(RequestsMock):
    """
    A stub/mock for the Strava API.

    This is a thin wrapper around `responses.RequestsMock`. It
    intercepts calls for registering mock responses and replaces these
    by a decorated method that supports convenience arguments that are
    not in the `responses` package.

    The methods `delete`, `get`, `head`, `options`, `patch`, `post`,
    and `put` are intercepted (and decorated), while the generic method
    `add` can be used to bypass the decoration and directly use the
    `responses` API.
    """

    def __getattribute__(self, name):
        attr = super().__getattribute__(name)
        if name in [
            "delete",
            "get",
            "head",
            "options",
            "patch",
            "post",
            "put",
        ]:
            return _api_method_adapter(attr)
        else:
            return attr
